#include <stdio.h>
#include <unistd.h>
#include <getopt.h>
#include <string.h>
#include <limits.h>
#include <errno.h>
#include <dirent.h>
#include <pwd.h>
#include <fcntl.h>

#include <sys/types.h>
#include <sys/param.h>
#include <sys/stat.h>

#include "config.h"

#define TITLE   "Safe RM (trashcan)"
#define VERSION  1.1
#define COPYRIGHT "Copyright (C) Ahmed Masud 1996"

#ifdef USES_CURSES
#define OPTIONS_STRING "dfirvRc"
#else
#define OPTIONS_STRING "dfirvR"
#endif

#define PN_RM          "rm"
#define PN_DESTROY     "destroy"
#define PN_RESTORE     "restore"
#define PN_PURGE       "purge"
#define PN_LT          "lt"

#define FATAL(arg,msg) { if (arg) { \
    fprintf (stderr, "%s: %s\nTry `%s --help' %s\n", argv[0], \
    msg, argv[0], "for more information"); exit(1); }  }

/* Following options are identical to the regular rm commands in terms of
 * their effect except that if the filename 'rm' it saves the contents first.
 */

const unsigned short opt_directory = 0x0001;
const unsigned short opt_force     = 0x0002;
const unsigned short opt_interact  = 0x0004;
const unsigned short opt_recurse   = 0x0008;
const unsigned short opt_verbose   = 0x0010;


#define OPT_HELP       1
#define OPT_VERSION    2
#define OPT_DESTROY    3
#define OPT_RESTORE    4
#define OPT_PURGE      5

const unsigned short opt_destroy   = 0x0020;
const unsigned short opt_restore   = 0x0040;
const unsigned short opt_purge     = 0x0080;
const unsigned short opt_remove    = 0x0100;
const unsigned short opt_overwrite = 0x0200;

#define OPT_BASEOPTS   0x001F

#define TRASHCAN_PATHS    "paths"
#define TRASHCAN_FILES    "files"
#define TRASHCAN_TEMPL    "trash.XXXXXX"

#ifdef MAXPATHLEN 
#undef MAXPATHLEN
#endif

#define MAXPATHLEN 255

#ifdef USES_CURSES
#define OPT_CURSES     'c'
const unsigned short opt_curses    = 0x1000;
#endif

/* 
 * Global variables
 */
static char progname[32];
static unsigned filecount = 0;
  


int destroyfiles(int argc, char **argv, int options);
int destroyfile(const char *file, int options);
int destroydir(const char *dir, int options);

int removefiles(int argc, char **argv, int options);
int removefile(const char *file, int options,
	       const char *trashcan_paths, const char *trashcan_files);
int removedir(const char *dir, int options,
	      const char *trashcan_paths, const char *trashcan_files);

int restorefiles(int options);
int restorefile(const char *file, int options, 
		const char *trashcan_paths, const char *trashcan_files);
int restoredir(const char *dir, int mode, int options);


int purgetrashcan(int options);
int purgefile(const char *file, int options,
	      const char *trashcan_paths, const char *trashcan_files);

int showtrash(void);
void showhelp(const char *progname);
void showversion(void);

int main(int argc, char **argv, char **argp) {
  int c;
  int options;
  int optidx = 0;

  char *pn;  // pointer to program name in argv[0];
  char **files;
  
  options = 0;
  
  pn = strrchr(argv[0], '/'); /* where is the last slash in the name? */
  
  if ( pn ) 
    strcpy(progname, pn+1);
  else
    strcpy(progname, argv[0]);
  
/* if (!strcmp(progname, PN_RM)) {  This comment is here for sake of 
 *   options |= 0;                  being complete.
 * }
 */
    
  if (!strcmp(progname, PN_DESTROY))
    options |= opt_destroy;
  else if (!strcmp(progname, PN_RESTORE))
    options |= opt_restore;
  else if (!strcmp(progname, PN_PURGE))
    options |= opt_purge;
  else
    options |= opt_remove;
  
  for ( ;; ) {
    
    static struct option long_options[] = 
      {
	/*
	 * Regular rm options.
	 */
	
	  { "directory", 0, 0, 'd' },
	  { "force", 0, 0, 'f' },
	  { "interactive", 0, 0, 'i' },
	  { "recursive", 0, 0, 'r' },
	  { "verbose", 0, 0, 'v' },
	  { "help", 0, 0, OPT_HELP },
	  { "version", 0, 0, OPT_VERSION },

	/*
	 * new safe-rm specific options
	 */
	
	  { "destroy", 0, 0, OPT_DESTROY }, /* destroy the files */
	  { "restore", 0, 0, OPT_RESTORE }, /* restore deleted files */
	  { "purge", 0, 0, OPT_PURGE }, /* purge deleted files */
	  { "overwrite", 0, 0, 'o'},    /* overwrite while restoring */
#ifdef USES_CURSES
	  { "curses", 0, 0, OPT_CURSES },
#endif
	  { 0, 0, 0, 0 }
      };    
    
    c = getopt_long (argc, argv, OPTIONS_STRING, long_options, &optidx);

    if ( c == -1 )
      break;
    
    switch (c) {
      
    case 'd':
      /* 'd' is not available while purging or restore */
      FATAL(options & (OPT_PURGE | OPT_RESTORE), "-d not supported");
      options |= opt_directory;
      break;
      
    case 'f':
      options |= opt_force;
      options &= ~opt_interact;
      break;
      
    case 'i':
      options |= opt_interact;
      options &= ~opt_force;
      break;
      
    case 'r':
    case 'R':
      options |= opt_recurse;
      break;
      
    case 'v':
      options |= opt_verbose;
      break;
            
    case 'o':
      options |= opt_overwrite;
      break;
      
    case OPT_HELP:
      showhelp(argv[0]);
      return 0;
      break;
      
    case '?':
    case ':':
      if ( options & opt_verbose )
	showhelp(argv[0]);

      return -1;
      break;

#ifdef USES_CURSES
    case OPT_CURSES:
      FATAL(1, "curses support not available (yet)");
      break;
#endif
      
    case OPT_VERSION:
      showversion();
      break;
      
    case OPT_DESTROY:
      FATAL (options & opt_purge, "cannot use purge and destroy simeltaneously");
      FATAL (options & opt_restore, "cannot use restore and destroy simeltaneously");
	
      options &= OPT_BASEOPTS;
      options |= opt_destroy;      
      break;

    case OPT_RESTORE:
      FATAL (options & opt_destroy, "cannot use restore and destroy simeltaneously");
      FATAL (options & opt_purge, "cannot use restore and purge simeltaneously");
      options &= OPT_BASEOPTS;      
      options |= opt_restore;
      break;
      
    case OPT_PURGE:
      FATAL (options & opt_destroy, "cannot use purge and destroy simeltaneously");
      FATAL (options & opt_restore, "cannot use restore and purge simeltaneously");      

      options &= OPT_BASEOPTS;      
      options |= opt_purge;
      break;
            
    }
  }

  files = &argv[optind];

  if (!strcmp(progname, PN_LT)) /* lists the contents of the trashcan */
    return showtrash();
  
  if (options & opt_overwrite)
    FATAL(!(options & opt_restore), "overwrite can only be used while restoring");
  
  if ( options & opt_restore ) {
    return restorefiles(options | (options & opt_force ? 0: opt_interact));
  }
  
  if ( options & opt_purge ) {
    return purgetrashcan(options|(options & opt_force ? 0: opt_interact));
  }
  
  FATAL(optind >= argc, "too few arguments");  /* If no files were specified exit! */
  
  /* If we are not forcing a delete and asking for multiple files for
   * deletion, we better ask for it.
   */
  
  if(((argc - optind) > 1) && !(options & opt_force))
    options |= opt_interact;
 
  
  if ( options & opt_destroy )
    return destroyfiles(argc - optind, files, options);
  
  if ( options & opt_remove )
    return removefiles(argc - optind, files, options);
  
  return 0;
}

void showversion(void) {
  fprintf( stdout, "%s version %1.2f\n", TITLE, VERSION);
  exit(0); 
}


void showhelp(const char *progname) {
    
  if (!strcmp(progname, PN_DESTROY)) {
    printf("Usage: %s [OPTION]... [PATH]...\n\n", progname);  
    printf("   -d, --directory        unlink directory, even if non-empty (not supported!)\n");
    printf("   -f, --force            ignore nonexistent files, never prompt\n");
    printf("   -i, --interactive      prompt before any removal\n");
    printf("   -v, --verbose          explain what is being done\n");
    printf("   -r, -R, --recursive    remove the contents of directories recursively\n");   
    printf("       --help             display this help and exit\n");
    printf("       --version          output version information and exit\n\n");
    return ;
  }
  
  if (!strcmp(progname, PN_RESTORE)) {
    printf("Usage: %s [OPTIONS]\n\n", progname);      
    printf("   -f, --force            ignore existing files, never prompt\n");
    printf("   -i, --interactive      prompt before any replacement (default)\n");
    printf("   -o, --overwrite        overwritting existing files\n");
    printf("   -v, --verbose          explain what is being done\n");
    printf("       --help             display this help and exit\n");
    printf("       --version          output version information and exit\n\n");
    return ;
  }
  
  if (!strcmp(progname, PN_PURGE)) {
    printf("Usage: %s [OPTIONS]\n\n", progname);          
    printf("   -f, --force            force trashcan purge\n");
    printf("   -i, --interactive      prompt before purging a file or directory (default)\n");
    printf("   -v, --verbose          explain what is being done\n");
    printf("       --help             display this help and exit\n");
    printf("       --version          output version information and exit\n\n");
    return ;
  }
  
  if (!strcmp(progname, PN_LT)) {
    printf("Usage: %s [OPTIONS]\n\n", progname);
    printf("       --help             display this help and exit\n");
    printf("       --version          output version information and exit\n\n");
    return ;
  }
  
  printf("Usage: %s [OPTION]... [PATH]...\n\n", progname);  
  printf("   -d, --directory        unlink directory, even if non-empty (not supported!)\n");
  printf("   -f, --force            ignore nonexistent files, never prompt\n");
  printf("   -i, --interactive      prompt before any removal\n");
  printf("   -v, --verbose          explain what is being done\n");
  printf("   -r, -R, --recursive    remove the contents of directories recursively\n");
  printf("       --help             display this help and exit\n");
  printf("       --version          output version information and exit\n\n");  
  return;
}


/* 
 * following are the remove functions which move the files to appropriate 
 * trashcan directory.
 */

int removefiles(int argc, char **argv, int options)  {
  
  int retval = 0, i;
  char trashcan[MAXPATHLEN], 
  trashcan_paths[MAXPATHLEN],
  trashcan_files[MAXPATHLEN];
  
  struct passwd *pwd;
  struct stat sb;

  pwd = getpwuid(geteuid());
  
  stat(pwd->pw_dir, &sb);

  if((pwd->pw_uid == 0) || !S_ISDIR(sb.st_mode))
    sprintf(trashcan, "%s", SYS_TRASH);
  else 
    sprintf(trashcan, "%s/%s", pwd->pw_dir, TRASH);
  
  if ( stat(trashcan, &sb) < 0 ) {
    if ( errno == ENOENT ) {
      if ( mkdir ( trashcan, 0700 ) < 0 ) {
	fprintf(stderr, "%s: unable to create trashcan\n", progname);
	return -1;
      }
    }
    else {
      fprintf(stderr, "%s: unknown error accessing trashcan\n", progname);
      return -1;
    }
  }
  
  sprintf(trashcan_paths, "%s/%s", trashcan, TRASHCAN_PATHS);
  sprintf(trashcan_files, "%s/%s", trashcan, TRASHCAN_FILES);
  
  if ( mkdir(trashcan_paths, 0700) < 0 ) {
    if ( errno != EEXIST ) {
      perror(progname);
      return -1;
    }    
  }
  
  if ( mkdir(trashcan_files, 0700) < 0 ) {
    if ( errno != EEXIST ) {
      perror(progname);
      return -1;
    }    
  }
  
  for ( i = 0; i < argc; i++ ) {
    if ( argv[i][strlen(argv[i])-1] == '/' )
      argv[i][strlen(argv[i])-1] = 0;
    retval = removefile(argv[i], options, trashcan_paths, trashcan_files);
  }
  
  return retval;
  
}

int removefile(const char *file, int options, 
	       const char *trashcan_paths, const char *trashcan_files) {

  int infd, outfd, retval;
  char temp[MAXPATHLEN];
  char t_path[MAXPATHLEN];
  char t_file[MAXPATHLEN];
  char buf[8192];
  struct stat sb;
  char errtxt[MAXPATHLEN];
  sprintf(errtxt, "%s: %s", progname, file);
  
  sprintf(temp, "trash.%d.XXXXXX", ++filecount);
  
  if ( mktemp(temp) == NULL ) {
    if (options & opt_force)
      return 0;
    fprintf (stderr, "%s: unable to create trashcan entry\n", errtxt);
    return -1;
  }
  
  sprintf(t_path, "%s/%s", trashcan_paths, temp);
  sprintf(t_file, "%s/%s", trashcan_files, temp);
  
  if ( lstat (file, &sb) < 0 ) {
    if ( options & opt_force )
      return 0;
    perror(errtxt);
    return -1;
  }

  if ( S_ISDIR (sb.st_mode) ) {
    if ( options & opt_recurse )
      return removedir(file, options, trashcan_paths, trashcan_files);
    if ( options & opt_force )
      return 0;
    fprintf (stderr, "%s: %s is a directory\n", progname, file);
    return -1;
  }

  if ( options & opt_interact ) {
    fprintf(stderr, "%s: move `%s' to trashcan? ", progname, file);
    fgets(temp, sizeof(temp), stdin);
    if (temp[0] != 'y') 
      return 0;
  }
  
  infd = open(file, O_RDONLY);
  
  if ( infd < 0 ) {
    if (options & opt_force)
      return 0;
    perror(errtxt);
    return -1;
  }
  
  outfd = open(t_file, O_WRONLY|O_CREAT, 0600);
  
  if ( outfd < 0 ) {
    if (options & opt_force)
      return 0;
    perror(errtxt);
    return -1;
  }

  
  while (1) {
    
    retval = read(infd, buf, sizeof(buf));
    if ( retval < 0 ) {
      unlink(t_file);    
      if (options & opt_force)
	return 0;
      perror(errtxt);
      return -1;
    }
      
    if ( !retval )
      break;
    
    write(outfd, buf, retval);
    
  }
  
  close(infd);
  close(outfd);
  
  outfd = open(t_path, O_WRONLY | O_CREAT, 0600);
  if (outfd < 0) {
    unlink(t_file);    
    if ( options & opt_force )
      return 0;
    perror(errtxt);
    return -1;
  }
  
  realpath(file, temp);
  write(outfd, temp, strlen(temp));
  sprintf(temp, " %u\n", sb.st_mode);
  write(outfd, temp, strlen(temp));
  close(outfd);
  
  if ( options & opt_verbose )
    fprintf(stderr, "%s: removing %s\n", progname, file);

  destroyfile(file, (options | opt_force) &  ~(opt_verbose | opt_interact));
  return 0;
}

int removedir(const char *file, int options,
	      const char *trash_paths, const char *trash_files)  {
  
  int outfd;
  char temp[MAXPATHLEN];
  char t_path[MAXPATHLEN];
  char errtxt[MAXPATHLEN];
  struct stat sb;
  struct dirent *d_ent;
  DIR *dir;
  
  if ( !strcmp (file, ".") || !strcmp (file, ".."))
    return 0;

  sprintf(errtxt, "%s: %s", progname, file);
  sprintf(temp, "trash.%02d.XXXXXX", ++filecount);  

  if ( mktemp(temp) == NULL ) {
    if (options & opt_force)
      return 0;
    fprintf (stderr, "%s: unable to create trashcan entry\n", errtxt);
    return -1;
  }
  
  sprintf(t_path, "%s/%s", trash_paths, temp);

  outfd  = open(t_path, O_WRONLY | O_CREAT, 0700);
    
  if ( outfd < 0 ) {
    if ( options & opt_force )
      return 0;
    perror(errtxt);
    return -1;
  }
  
  realpath(file, temp);
  write (outfd, temp, strlen (temp));
  lstat(file, &sb);
  sprintf(temp, " %u\n", sb.st_mode);
  write(outfd, temp, strlen(temp));
  close(outfd);
  
  if ( options & opt_interact ) {
    fprintf(stderr, "%s: descend directory `%s'? ", progname, file);
    fgets(temp, sizeof(temp), stdin);
    if (temp[0] != 'y') 
      return 0;
  }
  
  dir = opendir(file);
  do {
    d_ent = readdir(dir);
    if(d_ent == NULL)
      break;
    if(!strcmp(d_ent->d_name, ".") || !strcmp(d_ent->d_name, ".."))
      continue;
    /* temp now will contain the dir/file */
    sprintf(temp, "%s/%s", file, d_ent->d_name);
    removefile(temp, options, trash_paths, trash_files);
    
  } while(d_ent);

  if ( options & opt_interact ) {
    fprintf(stderr, "%s: move `%s' to trashcan? ", progname, file);
    fgets(temp, sizeof(temp), stdin);
    if (temp[0] != 'y') 
      return 0;
  }  
  
  destroydir(file, (options | opt_force) & ~(opt_verbose | opt_interact));
  if ( options & opt_verbose )
    fprintf (stderr, "%s removed\n", file);
  
  return 0;
}

/* Restore removed files */

int restorefiles(int options) {
  
  struct stat sb;
  char trashcan[MAXPATHLEN],
  trashcan_paths[MAXPATHLEN],
  trashcan_files[MAXPATHLEN];  
  struct passwd *pwd;  
  struct dirent *d_ent;
  DIR *dir;
  
  pwd = getpwuid(geteuid());

  stat(pwd->pw_dir, &sb);
  
  if((pwd->pw_uid == 0) || !S_ISDIR(sb.st_mode))
    sprintf(trashcan, "%s", SYS_TRASH);
  else 
    sprintf(trashcan, "%s/%s", pwd->pw_dir, TRASH);
  
  sprintf(trashcan_paths, "%s/%s", trashcan, TRASHCAN_PATHS);
  sprintf(trashcan_files, "%s/%s", trashcan, TRASHCAN_FILES);
  
  dir = opendir ( trashcan_paths );  /* Open the trashcan. */
  
  do {
    d_ent = readdir(dir);
    if (d_ent == NULL)
      break;
    if (!strcmp (d_ent->d_name, ".") || !strcmp (d_ent->d_name, ".."))
      continue;
    restorefile(d_ent->d_name, options, trashcan_paths, trashcan_files);
  } while (d_ent != NULL);
  
  return 0;
}

int restorefile(const char *file, int options, 
		const char *trashcan_paths, const char *trashcan_files) {
  
  int fdin, fdout;  
  FILE *fp;
  char buf[8192], 
  trash_file[MAXPATHLEN],
  trash_path[MAXPATHLEN],
  filename[MAXPATHLEN], 
  dirname[MAXPATHLEN], *p;
  struct stat sb;
  
  unsigned filemode;
  
  sprintf(trash_path, "%s/%s", trashcan_paths, file);
  sprintf(trash_file, "%s/%s", trashcan_files, file);
  
  fp = fopen(trash_path, "r");
  fscanf(fp, "%s %u\n", filename, &filemode);
  fclose(fp);
  
  if ( S_ISDIR(filemode) ) {
    int retval = restoredir(filename, filemode, options);
    destroyfile(trash_path, opt_force);     
    return retval;
  }
  
  p = strrchr(filename, '/');
  if ( p != filename ) { /* if it is not the first slash */
    strncpy(dirname, filename, p-filename);
    if (restoredir(dirname, options, 0755) < 0)
      return -1;
  }
  
  fdin = open (trash_file, O_RDONLY);
  
  if (stat (filename, &sb) == 0) { // Skip existing file;
    
    if (!(options & opt_overwrite)) {
      if (options & opt_force) {
	if (options & opt_verbose)
	  fprintf(stderr, "%s: skipping file `%s'\n",
		  progname, filename);
	close(fdin);
	return 0;
      }      
    }
    
    if (options & opt_interact) {
      fprintf(stderr, "%s: overwrite existing file `%s' ? ", 
	      progname, filename);
      fgets(buf, sizeof(buf), stdin);
      if(buf[0] != 'y') {
	if (options & opt_verbose)
	  fprintf(stderr, "%s: skipping file `%s'\n",
		  progname, filename);
	return -1;
      }
    }    
  }
  
  if ( options & opt_verbose ) {
    fprintf(stderr, "%s: replacing existing `%s' with trashcan copy\n",
	    progname, filename);
  }
  
  fdout = open (filename, O_WRONLY | O_CREAT, filemode);
  
  do {
    
    int retval;
    retval = read(fdin, buf, sizeof(buf));
    if (retval == 0) /* eof */
      break;
    write (fdout, buf, retval);
    
  } while(1);
  
  close(fdout);
  close(fdin);
  
  if ( options & opt_verbose )
    fprintf (stderr, "%s: restoring %s\n", progname, filename);
  
  destroyfile(trash_path, opt_force);
  destroyfile(trash_file, opt_force);
  
  return 0;
}

int restoredir(const char *file, int mode, int options) {
  
  struct stat sb;
  char temp[MAXPATHLEN];
  char *p1, *p2;
  char errtxt[32 + MAXPATHLEN];
  sprintf(errtxt, "%s: mkdir ", progname);
  
  p1 = (char *) file;
  
  do {
    
    int done = 0;
    p2 = strchr(p1+1, '/'); /* +1 skips leading slash */    
    
    if (p2 == NULL) {
      p2 = p1+strlen(p1);
      done = 1;
    }
    
    memset(temp, 0, sizeof(temp));
    strncpy ( temp, file, p2-file );
    
    if ( stat(temp, &sb) < 0 ) {
      if ( errno == ENOENT ) { /* file does not exist */
	if ( options & opt_verbose )
	  fprintf (stderr, "%s: restoring %s\n", progname, temp);
	if ( mkdir ( temp, 0700 ) < 0 ) {
	  if ( options & opt_force )
	    return -1;
	  perror (errtxt); 
	  return -1; /* Unable to create a parent directory */
	}
	/* make directory was successful so do nothing and just go on */
      }
      else {
	if ( options & opt_force )
	  return -1;
	perror(errtxt);
	return -1;
      }
    }
    
    if ( done )
      break;
    
    p1 = p2; /* we are on to the next portion */
    
  } while (p2 != NULL);
  
  return 0;
}

/* purge the trashcan */

int purgetrashcan(int options) {
  
  struct stat sb;
  char trashcan[MAXPATHLEN],
  trashcan_paths[MAXPATHLEN],
  trashcan_files[MAXPATHLEN];  
  struct passwd *pwd;  
  struct dirent *d_ent;
  DIR *dir;
  
  pwd = getpwuid(geteuid());

  stat(pwd->pw_dir, &sb);
  
  if((pwd->pw_uid == 0) || !S_ISDIR(sb.st_mode))
    sprintf(trashcan, "%s", SYS_TRASH);
  else 
    sprintf(trashcan, "%s/%s", pwd->pw_dir, TRASH);
  
  sprintf(trashcan_paths, "%s/%s", trashcan, TRASHCAN_PATHS);
  sprintf(trashcan_files, "%s/%s", trashcan, TRASHCAN_FILES);

  dir = opendir ( trashcan_paths );  /* Open the trashcan. */
  
  do {
    d_ent = readdir(dir);
    if (d_ent == NULL)
      break;
    if (!strcmp (d_ent->d_name, ".") || !strcmp (d_ent->d_name, ".."))
      continue;    
    purgefile(d_ent->d_name, options, trashcan_paths, trashcan_files);
  } while (d_ent != NULL);
  
  return 0;
}

int purgefile(const char *file, int options, 
	      const char *trashcan_paths, const char *trashcan_files) {
  
  char trash_path[MAXPATHLEN];
  char trash_file[MAXPATHLEN];
  char filename[MAXPATHLEN];
  
  FILE *fp;
  unsigned mode;
  
  sprintf(trash_path, "%s/%s", trashcan_paths, file);  
  sprintf(trash_file, "%s/%s", trashcan_files, file);
  
  fp = fopen(trash_path, "r");
  fscanf (fp, "%s %u", filename, &mode);
  fclose(fp);
  
  if ( options & opt_interact ) {
    char buf[32];
    fprintf(stderr, "%s: remove `%s' from trashcan? ",
	    progname, filename);  
    fgets(buf, sizeof(buf), stdin);
    if ( buf[0] != 'y' ) {
      if ( options & opt_verbose ) 
	fprintf(stderr, "%s: skipping `%s'\n", progname, filename);
      return -1;
    }
  }
  
  if ( options & opt_verbose )
    fprintf(stderr, "%s: purging `%s'\n", progname, filename);
  
  unlink(trash_path);
  unlink(trash_file);
  
  return 0;
  
}

int showtrash(void) {
  
  struct stat sb;
  char trashcan[MAXPATHLEN],
  trashcan_paths[MAXPATHLEN],
  trashcan_files[MAXPATHLEN];  
  struct passwd *pwd;  
  struct dirent *d_ent;
  DIR *dir;
  
  pwd = getpwuid(geteuid());

  stat(pwd->pw_dir, &sb);
  
  if((pwd->pw_uid == 0) || !S_ISDIR(sb.st_mode))
    sprintf(trashcan, "%s", SYS_TRASH);
  else 
    sprintf(trashcan, "%s/%s", pwd->pw_dir, TRASH);
  
  sprintf(trashcan_paths, "%s/%s", trashcan, TRASHCAN_PATHS);
  sprintf(trashcan_files, "%s/%s", trashcan, TRASHCAN_FILES);

  dir = opendir ( trashcan_paths );  /* Open the trashcan. */
  
  do {
    int mode;
    FILE *fp;
    char filename[MAXPATHLEN];
    char trash_path[MAXPATHLEN];
    
    d_ent = readdir(dir);
    
    if (d_ent == NULL)
      break;
    
    if (!strcmp (d_ent->d_name, ".") || !strcmp (d_ent->d_name, ".."))
      continue;
    
    sprintf(trash_path, "%s/%s", trashcan_paths, d_ent->d_name);

    fp = fopen(trash_path, "r");
    memset(filename, 0, sizeof(filename));
    fscanf(fp, "%s %u\n", filename, &mode);
    fclose(fp);
    
    printf("%s\n", filename);
    
  } while (d_ent != NULL);
  
  return 0;
}
/* following destroy functions loose the files permanently! */

int destroyfiles(int argc, char **argv, int options) {  
  int retval = 0, i;
  for ( i = 0; i < argc; i++ ) {
    if ( argv[i][strlen(argv[i])-1] == '/' )
      argv[i][strlen(argv[i])-1] = 0;
    retval = destroyfile(argv[i], options);
  }  
  return retval;
}

int destroydir(const char *file, int options) {
  DIR *dir;
  struct dirent *d_ent;
  char subfile[MAXPATHLEN];

  if ( !strcmp (file, ".") || !strcmp (file, ".."))
    return 0;
  
  if (rmdir (file) < 0 ) {
    
    if(errno != ENOTEMPTY)
      return -1;

    if (!(options & opt_force)) {
      options |= opt_interact;
      fprintf(stderr, "%s: descend directory %s? ", progname, file);
      /* we can use subfile as a temporary for now! */
      fgets(subfile, sizeof(subfile), stdin);
      if (subfile[0] != 'y') {
	return 0;
      }
    }
    
    dir = opendir(file);
  
    if ( dir == NULL )
      return -1;
    
    /* If we are not forcing it is better to ask! */
    if (!(options & opt_force)) 
      options |= opt_recurse;
    
    do {
      d_ent = readdir(dir);
      if ( d_ent == NULL )
	break;
      if (!strcmp(d_ent->d_name, ".") || !strcmp(d_ent->d_name, ".."))
	continue;
      sprintf(subfile, "%s/%s", file, d_ent->d_name);
      destroyfile(subfile, options);
    } while(d_ent != NULL);
    
  closedir(dir);
    
    if (rmdir (file) < 0 )
      if(errno != ENOTEMPTY)
      return -1;
    
  }
  
  if ( options & opt_verbose )
    fprintf ( stderr, "%s\n", file);
  
  return 0;
}


int destroyfile(const char *file, int options) {
  struct stat fs;
  int retval;
  char errstr[32+MAXPATHLEN];
  char yn[16];
  
  retval = lstat(file, &fs);
  
  sprintf(errstr, "%s: %s", progname, file);
  
  if (retval < 0)  {
    if (options & opt_force) {
      return 0;
    }
    perror(errstr);
    return -1;
  }

  if (S_ISDIR(fs.st_mode)) {
    if (options & opt_recurse) {
      if ( destroydir(file, options) < 0) {
	perror(errstr);
	return -1;
      }
      return 0;
    }
    fprintf(stderr, "%s: is a directory\n", errstr);
    return -1;
  }
  else  {
    if ( options & opt_interact ) {
      fprintf(stderr, "%s: remove %s ? ", progname, file);
      fgets(yn, sizeof(yn), stdin);      
      if ( yn[0] != 'y' )
	return 0;      
    }
    
    if ( unlink (file) < 0 ) {
      if ( options & opt_force )
	return 0;	 
      perror(errstr);
      return -1;
    }
  }
  if ( options & opt_verbose )
    fprintf ( stderr, "%s\n", file);
  
  return 0;
}
